import json
import logging
import os
from dataclasses import dataclass

import requests
from KicadModTree import Footprint, KicadFileHandler, Pad, Text, Translation

from .footprint_handlers import handlers, mil2mm


@dataclass
class FootprintInfo:
    max_X: float = -10000
    max_Y: float = -10000
    min_X: float = 10000
    min_Y: float = 10000
    footprint_name: str = ""
    output_dir: str = ""
    footprint_lib: str = ""
    model_base_variable: str = ""
    model_dir: str = ""
    origin: tuple = (0, 0)
    models: str = ""


def create_footprint(
    footprint_component_uuid,
    component_id,
    footprint_lib,
    output_dir,
    model_base_variable,
    model_dir,
    skip_existing,
    models,
):
    logging.info("Creating footprint ...")

    (
        footprint_name,
        datasheet_link,
        footprint_shape,
        translation,
    ) = get_footprint_info(footprint_component_uuid)

    if skip_existing and os.path.isfile(
        os.path.join(output_dir, footprint_lib, footprint_name + ".kicad_mod")
    ):
        logging.info(f"Footprint {footprint_name} already exists, skipping.")
        return f"{footprint_lib}:{footprint_name}", datasheet_link

    # init kicad footprint
    kicad_mod = Footprint(f'"{footprint_name}"')
    kicad_mod.setDescription(f"{footprint_name} footprint")  # TODO Set real description
    kicad_mod.setTags(f"{footprint_name} footprint {component_id}")

    footprint_info = FootprintInfo(
        footprint_name=footprint_name,
        output_dir=output_dir,
        footprint_lib=footprint_lib,
        model_base_variable=model_base_variable,
        model_dir=model_dir,
        origin=translation,
        models=models,
    )

    # for each line in data : use the appropriate handler
    for line in footprint_shape:
        args = [
            i for i in line.split("~") if i
        ]  # split and remove empty string in list
        model = args[0]
        logging.debug(args)
        if model not in handlers:
            logging.warning(f"footprint : model not in handler :  {model}")
        else:
            handlers.get(model)(args[1:], kicad_mod, footprint_info)

    if any(
        isinstance(child, Pad) and child.type == Pad.TYPE_THT
        for child in kicad_mod.getAllChilds()
    ):
        kicad_mod.setAttribute("through_hole")
    else:
        kicad_mod.setAttribute("smd")

    kicad_mod.insert(Translation(-mil2mm(translation[0]), -mil2mm(translation[1])))

    # Translate the footprint max and min values to the origin
    footprint_info.max_X -= mil2mm(translation[0])
    footprint_info.max_Y -= mil2mm(translation[1])
    footprint_info.min_X -= mil2mm(translation[0])
    footprint_info.min_Y -= mil2mm(translation[1])

    # set general values
    kicad_mod.append(
        Text(
            type="reference",
            text="REF**",
            at=[
                (footprint_info.min_X + footprint_info.max_X) / 2,
                footprint_info.min_Y - 2,
            ],
            layer="F.SilkS",
        )
    )
    kicad_mod.append(
        Text(
            type="user",
            text="${REFERENCE}",
            at=[
                (footprint_info.min_X + footprint_info.max_X) / 2,
                (footprint_info.min_Y + footprint_info.max_Y) / 2,
            ],
            layer="F.Fab",
        )
    )
    kicad_mod.append(
        Text(
            type="value",
            text=footprint_name,
            at=[
                (footprint_info.min_X + footprint_info.max_X) / 2,
                footprint_info.max_Y + 2,
            ],
            layer="F.Fab",
        )
    )

    if not os.path.exists(f"{output_dir}/{footprint_lib}"):
        os.makedirs(f"{output_dir}/{footprint_lib}")

    # output kicad model
    file_handler = KicadFileHandler(kicad_mod)
    file_handler.writeFile(f"{output_dir}/{footprint_lib}/{footprint_name}.kicad_mod")
    logging.info(f"Created '{output_dir}/{footprint_lib}/{footprint_name}.kicad_mod'")

    # return the datasheet link and footprint name to be linked with the symbol
    return (f"{footprint_lib}:{footprint_name}", datasheet_link)


def get_footprint_info(footprint_component_uuid):
    # fetch the component data from easyeda library
    response = requests.get(
        f"https://easyeda.com/api/components/{footprint_component_uuid}"
    )

    if response.status_code == requests.codes.ok:
        data = json.loads(response.content.decode())
    else:
        logging.error(
            "create_footprint error. Requests returned with error code "
            f"{response.status_code}"
        )
        return ("", None, "", (0, 0))

    footprint_shape = data["result"]["dataStr"]["shape"]
    x = data["result"]["dataStr"]["head"]["x"]
    y = data["result"]["dataStr"]["head"]["y"]
    try:
        datasheet_link = data["result"]["dataStr"]["head"]["c_para"]["link"]
    except KeyError:
        datasheet_link = ""
        logging.warning("Could not retrieve datasheet link from EASYEDA")

    footprint_name = (
        data["result"]["title"]
        .replace(" ", "_")
        .replace("/", "_")
        .replace("(", "_")
        .replace(")", "_")
    )

    if not footprint_name:
        footprint_name = "NoName"
        logging.warning(
            "Could not retrieve components information from EASYEDA, default name "
            "'NoName'."
        )

    return (footprint_name, datasheet_link, footprint_shape, (x, y))



import logging
import re

RELATIVE_OFFSET = 0.254
ABSOLUTE_OFFSET_X = 101.6
ABSOLUTE_OFFSET_Y = -63.5

__all__ = ["handlers", "h_R", "h_E", "h_P", "h_T", "h_PL", "h_PG", "h_PT", "h_A"]


def mil2mm(data):
    return float(data) / 3.937


def h_R(data, translation, kicad_symbol):
    """
    Rectangle handler
    """

    try:
        if len(data) == 12:
            X1 = float(data[0])
            Y1 = float(data[1])
            X2 = float(X1) + float(data[4])
            Y2 = float(Y1) + float(data[5])
        else:
            X1 = float(data[0])
            Y1 = float(data[1])
            X2 = float(X1) + float(data[2])
            Y2 = float(Y1) + float(data[3])

        X1 = mil2mm(X1 - translation[0])
        Y1 = -mil2mm(Y1 - translation[1])
        X2 = mil2mm(X2 - translation[0])
        Y2 = -mil2mm(Y2 - translation[1])

        kicad_symbol.drawing += f"""
      (rectangle
        (start {X1} {Y1})
        (end {X2} {Y2})
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type background))
      )"""
    except Exception as e:
        print(e)
        logging.error("symbol : failed to add a rectangle")


def h_E(data, translation, kicad_symbol):
    """
    Circle
    """

    try:
        X1 = mil2mm(float(data[0]) - translation[0])
        Y1 = -mil2mm(float(data[1]) - translation[1])
        radius = mil2mm(float(data[2]))

        kicad_symbol.drawing += f"""
      (circle
        (center {X1} {Y1})
        (radius {radius})
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type background))
      )"""
    except Exception as e:
        print(e)
        logging.error("symbol : failed to add circle")


def h_P(data, translation, kicad_symbol):
    """
    Add Pin to the symbol
    """

    if len(data) == 24:  # sometimes, the rotation parameter is not in the list.
        data.insert(5, "0")
    elif len(data) == 28:
        data.insert(1, "0")

    if data[1] == "0":
        electrical_type = "unspecified"
    elif data[1] == "1":
        electrical_type = "input"
    elif data[1] == "2":
        electrical_type = "output"
    elif data[1] == "3":
        electrical_type = "bidirectional"
    elif data[1] == "4":
        electrical_type = "power_in"
    else:
        electrical_type = "unspecified"

    pinNumber = data[2]
    pinName = data[13]

    X = round(mil2mm(float(data[3]) - translation[0]), 3)
    Y = round(-mil2mm(float(data[4]) - translation[1]), 3)

    if data[5] in ["0", "90", "180", "270"]:
        rotation = (int(data[5]) + 180) % 360
    else:
        rotation = 0
        logging.warning(
            f'symbol : pin number {pinNumber} : "{pinName}" failed to find orientation.'
            "Using Default orientation"
        )

    if rotation == 0 or rotation == 180:
        length = round(mil2mm(abs(float(data[8].split("h")[-1]))), 3)
    elif rotation == 90 or rotation == 270:
        length = mil2mm(abs(float(data[8].split("v")[-1])))
    else:
        length = 2.54
        logging.warning(
            f'symbol : pin number {pinNumber} : "{pinName}" failed to find length.'
            "Using Default length"
        )

    try:
        # If on pin name/number is not hidden, show set the synmbol hide property to 0
        if data[9].split("^^")[1] != "0":
            kicad_symbol.pinNamesHide = ""
        if data[17].split("^^")[1] != "0":
            kicad_symbol.pinNumbersHide = ""
    except Exception:
        kicad_symbol.pinNamesHide = ""
        kicad_symbol.pinNumbersHide = ""

    try:
        nameSize = mil2mm(float(data[16].replace("pt", "")))
        numberSize = mil2mm(float(data[24].replace("pt", "")))
    except Exception:
        nameSize = 1
        numberSize = 1

    kicad_symbol.drawing += f"""
      (pin {electrical_type} line
        (at {X} {Y} {rotation})
        (length {length})
        (name "{pinName}" (effects (font (size {nameSize} {nameSize}))))
        (number "{pinNumber}" (effects (font (size {numberSize} {numberSize}))))
      )"""


def h_T(data, translation, kicad_symbol):
    """
    Annotation handler
    """

    try:
        X = mil2mm(float(data[1]) - translation[0])
        Y = -mil2mm(float(data[2]) - translation[1])
        angle = (float(data[3]) * 10 + 1800) % 3600

        fontSize = mil2mm(float(data[6].replace("pt", "")))

        text = data[10]
        kicad_symbol.drawing += f"""
      (text
        "{text}"
        (at {X} {Y} {angle})
        (effects (font (size {fontSize} {fontSize})))
      )"""
    except Exception:
        logging.error("failed to add text to symbol")


def h_PL(data, translation, kicad_symbol):
    """
    Polygone handler
    """

    try:
        pathString = data[0].split(" ")
        polypts = []
        for i, _ in enumerate(pathString[::2]):
            polypts.append(
                f"(xy {mil2mm(float(pathString[2 * i]) - translation[0])} "
                f"{-mil2mm(float(pathString[2 * i + 1]) - translation[-1])})"
            )
        polystr = "\n          ".join(polypts)

        kicad_symbol.drawing += f"""
      (polyline
        (pts
          {polystr}
        )
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type none))
      )"""
    except Exception:
        logging.error("symbol : failed to add a polygone")


def h_PG(data, translation, kicad_symbol):
    """
    Closed polygone handler
    """

    try:
        pathString = [i for i in data[0].split(" ") if i]
        polypts = []
        for i, _ in enumerate(pathString[::2]):
            polypts.append(
                f"(xy {mil2mm(float(pathString[2 * i]) - translation[0])} "
                f"{-mil2mm(float(pathString[2 * i + 1]) - translation[1])})"
            )
        polypts.append(polypts[0])
        polystr = "\n          ".join(polypts)

        kicad_symbol.drawing += f"""
      (polyline
        (pts
          {polystr}
        )
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type background))
      )"""
    except Exception:
        logging.error("symbol : failed to add a polygone")


def h_PT(data, translation, kicad_symbol):
    """
    Triangle handler
    """

    try:
        data[0] = data[0].replace("M ", "").replace("L ", "").replace(" Z", "")
        h_PG(data, translation, kicad_symbol)
    except Exception:
        logging.error("symbol : failed to add a triangle")


def h_A(data, translation, kicad_symbol):
    """
    Arc handler
    """

    from math import acos, cos, pi, sin, sqrt

    # ruff: disable [E741]
    # Function reversed from https://easyeda.com/editor/6.5.5/js/editorPCB.min.js
    def getCenterParam(match):
        e = float([i for i in re.split(r" |,", match[0][1]) if i][0])
        t = float([i for i in re.split(r" |,", match[0][1]) if i][1])
        s = float([i for i in re.split(r" |,", match[1][1]) if i][0])
        l = float([i for i in re.split(r" |,", match[1][1]) if i][1])
        r = float([i for i in re.split(r" |,", match[1][1]) if i][3])
        o = float([i for i in re.split(r" |,", match[1][1]) if i][4])
        n = float([i for i in re.split(r" |,", match[1][1]) if i][5])
        a = float([i for i in re.split(r" |,", match[1][1]) if i][6])

        def c(e, t, n, a):
            i = e * n + t * a
            r = sqrt((e * e + t * t) * (n * n + a * a))
            o = acos(i / r)
            return o

        f = 2 * pi
        if o < 0:
            o = -o
        if s < 0:
            s = -s
        if o == s:
            l = 0
        C = sin(l)
        y = cos(l)
        b = (e - n) / 2
        v = (t - a) / 2
        S = (e + n) / 2
        P = (t + a) / 2
        if o < 0.00001 or s < 0.00001:
            h = c(1, 0, n - e, a - t)
            return (S, P, h, pi)
        A = y * b + C * v
        T = y * v - C * b
        D = A * A / (o * o) + T * T / (s * s)
        if D > 1:
            o *= sqrt(D)
            s *= sqrt(D)
        k = o * s
        M = o * T
        I = s * A
        L = M * M + I * I
        if not L:
            return (S, P, 0, 0)
        w = (k * k - L) / L
        w = sqrt(abs(w))
        O = w * M / s
        R = -w * I / o
        u = y * O - C * R + S
        g = C * O + y * R + P
        E = (A - O) / o
        N = (A + O) / o
        F = (T - R) / s
        x = (T + R) / s
        h = c(1, 0, E, F)
        m = c(E, F, -N, -x)
        while m > f:
            m -= f
        while m < 0:
            m += f
        if r != 0:
            m -= f
        return (u, g, h, m)

    # ruff: enable [E741]

    try:
        match = re.findall(r"([MA])([eE ,\-\+.\d]+)", data[0])
        cx, cy, theta, deltaTheta = getCenterParam(match)
        radius = float([i for i in re.split(r" |,", match[1][1]) if i][0])
        theta /= 2
        Xstart = cx + radius * cos(theta)
        Ystart = -(cy - radius * sin(theta))
        Xend = cx + radius * cos(theta + deltaTheta)
        Yend = -(cy - radius * sin(theta + deltaTheta))
        Xmid = cx + radius * cos(theta + deltaTheta / 2)
        Ymid = -(cy - radius * sin(theta + deltaTheta / 2))

        Xstart = mil2mm(Xstart - translation[0])
        Ystart = -mil2mm(Ystart - translation[1])
        Xend = mil2mm(Xend - translation[0])
        Yend = -mil2mm(Yend - translation[1])
        Xmid = mil2mm(Xmid - translation[0])
        Ymid = -mil2mm(Ymid - translation[1])

        kicad_symbol.drawing += f"""
      (arc
        (start {Xstart} {Ystart})
        (mid {Xmid} {Ymid})
        (end {Xend} {Yend})
        (stroke (width 0) (type default) (color 0 0 0 0))
        (fill (type none))
      )"""
    except Exception:
        logging.error("symbol : failed to add an arc")


handlers = {
    "R": h_R,
    "E": h_E,
    "P": h_P,
    "T": h_T,
    "PL": h_PL,
    "PG": h_PG,
    "PT": h_PT,
    "A": h_A,
    # "J" : h_NotYetImplemented,
    # "N" : h_NotYetImplemented,
    # "BE" : h_NotYetImplemented,
    # "AR" : h_NotYetImplemented,
    # "O" : h_NotYetImplemented,
}



import logging
import os
import re

import requests
from KicadModTree import Model

wrl_header = """#VRML V2.0 utf8
#created by JLC2KiCad_lib using the JLCPCB library
#for more info see https://github.com/TousstNicolas/JLC2KICAD_lib
"""


def mil2mm(data):
    return float(data) / 3.937


def get_StepModel(
    component_uuid,
    footprint_info,
    kicad_mod,
    translationX,
    translationY,
    translationZ,
    rotation,
):
    logging.info("Downloading STEP Model ...")

    # `qAxj6KHrDKw4blvCG8QJPs7Y` is a constant in
    # https://modules.lceda.cn/smt-gl-engine/0.8.22.6032922c/smt-gl-engine.js
    # and points to the bucket containing the step files.

    response = requests.get(
        f"https://modules.easyeda.com/qAxj6KHrDKw4blvCG8QJPs7Y/{component_uuid}"
    )

    if response.status_code != requests.codes.ok:
        logging.error("request error, no Step model found")
        return

    ensure_footprint_lib_directories_exist(footprint_info)
    filename = (
        f"{footprint_info.output_dir}/"
        f"{footprint_info.footprint_lib}/"
        f"{footprint_info.model_dir}/"
        f"{footprint_info.footprint_name}.step"
    )
    with open(filename, "wb") as f:
        f.write(response.content)

    logging.info(f"STEP model created at {filename}")

    if footprint_info.model_base_variable:
        if footprint_info.model_base_variable.startswith("$"):
            path_name = (
                f'"{footprint_info.model_base_variable}/'
                f"{footprint_info.model_dir}/"
                f'{footprint_info.footprint_name}.step"'
            )
        else:
            path_name = (
                f'"$({footprint_info.model_base_variable})/'
                f"{footprint_info.model_dir}/"
                f'{footprint_info.footprint_name}.step"'
            )
    else:
        path_name = f"{footprint_info.model_dir}/{footprint_info.footprint_name}.step"

    translationX = (translationX - footprint_info.origin[0]) / 100
    translationY = -(translationY - footprint_info.origin[1]) / 100
    translationZ = float(translationZ) / 100

    kicad_mod.append(
        Model(
            filename=path_name,
            at=[translationX, translationY, translationZ],
            rotate=[-float(axis_rotation) for axis_rotation in rotation.split(",")],
        )
    )
    logging.info(f"added {path_name} to footprint")


def get_WrlModel(
    component_uuid,
    footprint_info,
    kicad_mod,
    translationX,
    translationY,
    translationZ,
    rotation,
):
    logging.info("Creating WRL model ...")

    response = requests.get(
        f"https://easyeda.com/analyzer/api/3dmodel/{component_uuid}"
    )
    if response.status_code == requests.codes.ok:
        text = response.content.decode()
    else:
        logging.error("request error, no 3D model found")
        return ()

    wrl_content = wrl_header

    # get material list
    pattern = "newmtl .*?endmtl"
    matchs = re.findall(pattern=pattern, string=text, flags=re.DOTALL)

    materials = {}
    for match in matchs:
        material = {}
        material_id = ""
        for value in match.split("\n"):
            if value[0:6] == "newmtl":
                material_id = value.split(" ")[1]
            elif value[0:2] == "Ka":
                material["ambientColor"] = value.split(" ")[1:]
            elif value[0:2] == "Kd":
                material["diffuseColor"] = value.split(" ")[1:]
            elif value[0:2] == "Ks":
                material["specularColor"] = value.split(" ")[1:]
            elif value[0] == "d":
                material["transparency"] = value.split(" ")[1]

        materials[material_id] = material

    # get vertices list
    pattern = "v (.*?)\n"
    matchs = re.findall(pattern=pattern, string=text, flags=re.DOTALL)

    vertices = []
    for vertice in matchs:
        vertices.append(
            " ".join(
                [str(round(float(coord) / 2.54, 4)) for coord in vertice.split(" ")]
            )
        )

    # get shape list
    shapes = text.split("usemtl")[1:]
    for shape in shapes:
        lines = shape.split("\n")
        material = materials[lines[0].replace(" ", "")]
        index_counter = 0
        link_dict = {}
        coordIndex = []
        points = []
        for line in lines[1:]:
            if len(line) > 0:
                face = [int(index) for index in line.replace("//", "").split(" ")[1:]]
                face_index = []
                for index in face:
                    if index not in link_dict:
                        link_dict[index] = index_counter
                        face_index.append(str(index_counter))
                        points.append(vertices[index - 1])
                        index_counter += 1
                    else:
                        face_index.append(str(link_dict[index]))
                face_index.append("-1")
                coordIndex.append(",".join(face_index) + ",")
        points.insert(-1, points[-1])

        shape_str = f"""
Shape{{
	appearance Appearance {{
		material  Material 	{{ 
			diffuseColor {" ".join(material["diffuseColor"])} 
			specularColor {" ".join(material["specularColor"])}
			ambientIntensity 0.2
			transparency {material["transparency"]}
			shininess 0.5
		}}
	}}
	geometry IndexedFaceSet {{
		ccw TRUE 
		solid FALSE
		coord DEF co Coordinate {{
			point [
				{(", ").join(points)}
			]
		}}
		coordIndex [
			{"".join(coordIndex)}
		]
	}}
}}"""

        wrl_content += shape_str

    ensure_footprint_lib_directories_exist(footprint_info)

    filename = (
        f"{footprint_info.output_dir}/"
        f"{footprint_info.footprint_lib}/"
        f"{footprint_info.model_dir}/"
        f"{footprint_info.footprint_name}.wrl"
    )
    with open(filename, "w") as f:
        f.write(wrl_content)

    if footprint_info.model_base_variable:
        if footprint_info.model_base_variable.startswith("$"):
            path_name = (
                f'"{footprint_info.model_base_variable}/'
                f"{footprint_info.model_dir}/"
                f'{footprint_info.footprint_name}.wrl"'
            )
        else:
            path_name = (
                f'"$({footprint_info.model_base_variable})/'
                f"{footprint_info.model_dir}/"
                f'{footprint_info.footprint_name}.wrl"'
            )
    else:
        path_name = f"{footprint_info.model_dir}/{footprint_info.footprint_name}.wrl"

    translationX = (translationX - footprint_info.origin[0]) / 100
    translationY = -(translationY - footprint_info.origin[1]) / 100
    translationZ = float(translationZ) / 100

    # Check if a model has already been added to the footprint to prevent duplicates
    if any(isinstance(child, Model) for child in kicad_mod.getAllChilds()):
        logging.info("WRL model created at {filename}")
        logging.info(
            "WRL model was not added to the footprint to prevent duplicates with STEP "
            "model"
        )
    else:
        kicad_mod.append(
            Model(
                filename=path_name,
                at=[translationX, translationY, translationZ],
                rotate=[-float(axis_rotation) for axis_rotation in rotation.split(",")],
            )
        )
        logging.info(f"added {path_name} to footprintc")


def ensure_footprint_lib_directories_exist(footprint_info):
    if not os.path.exists(
        f"{footprint_info.output_dir}/{footprint_info.footprint_lib}"
    ):
        os.makedirs(f"{footprint_info.output_dir}/{footprint_info.footprint_lib}")

    if not os.path.exists(
        f"{footprint_info.output_dir}/{footprint_info.footprint_lib}/{footprint_info.model_dir}"
    ):
        os.makedirs(
            f"{footprint_info.output_dir}/{footprint_info.footprint_lib}/{footprint_info.model_dir}"
        )



import logging
import sys


def set_logging(logging_level, logging_file):
    LOGGING_FILE = "JLC2KiCad_lib.log"

    if logging_file:
        logging.basicConfig(
            filename=LOGGING_FILE, format="%(asctime)s - %(levelname)s - %(message)s"
        )

    root_logger = logging.getLogger()
    root_logger.setLevel(logging_level)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    root_logger.addHandler(handler)
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
    handler.setFormatter(formatter)



import json
import logging
import re
from math import acos, pi, pow, sqrt

from KicadModTree import (
    Arc,
    Circle,
    Line,
    Pad,
    Polygon,
    RectFill,
    RectLine,
    Text,
    Vector2D,
)

from .model3d import get_StepModel, get_WrlModel

__all__ = [
    "handlers",
    "h_TRACK",
    "h_PAD",
    "h_ARC",
    "h_CIRCLE",
    "h_SOLIDREGION",
    "h_SVGNODE",
    "h_VIA",
    "h_RECT",
    "h_HOLE",
    "h_TEXT",
    "mil2mm",
]

layer_correspondance = {
    "1": "F.Cu",
    "2": "B.Cu",
    "3": "F.SilkS",
    "4": "B.Silks",
    "5": "F.Paste",
    "6": "B.Paste",
    "7": "F.Mask",
    "8": "B.Mask",
    "10": "Edge.Cuts",
    "12": "F.Fab",
    "99": "F.SilkS",
    "100": "F.SilkS",
    "101": "F.SilkS",
}


def mil2mm(data):
    return float(data) / 3.937


def h_TRACK(data, kicad_mod, footprint_info):
    width = mil2mm(data[0])

    # "S$xx" is sometimes inserted at index 2 ?
    if "$" in data[2]:
        points = [mil2mm(p) for p in data[3].split(" ") if p]
    else:
        points = [mil2mm(p) for p in data[2].split(" ") if p]

    for i in range(int(len(points) / 2) - 1):
        start = [points[2 * i], points[2 * i + 1]]
        end = [points[2 * i + 2], points[2 * i + 3]]
        try:
            layer = layer_correspondance[data[1]]
        except Exception:
            logging.exception("footprint h_TRACK: layer correspondance not found")
            layer = "F.SilkS"

        # update footprint borders
        footprint_info.max_X = max(footprint_info.max_X, start[0], end[0])
        footprint_info.min_X = min(footprint_info.min_X, start[0], end[0])
        footprint_info.max_Y = max(footprint_info.max_Y, start[1], end[1])
        footprint_info.min_Y = min(footprint_info.min_Y, start[1], end[1])

        # append line to kicad_mod
        kicad_mod.append(Line(start=start, end=end, width=width, layer=layer))


def h_PAD(data, kicad_mod, footprint_info):
    """
    Append a pad to the footprint

    data : [
        0 : shape type
        1 : pad position x
        2 : pad position y
        3 : pad size x
        4 : pad size y
        5 : layer
        6 : pad number
        7 : drill size
        8 : Polygon nodes "skipped for some shapes"
        9 : rotation
        10 :
        11 : drill offset
        12 :
        13 :
        14 :
        15 :
        16 :
        17 : ? position
    ]
    """

    # PAD layer definition
    TOPLAYER = "1"
    BOTTOMLAYER = "2"
    MULTILAYER = "11"

    shape_type = data[0]
    at = [mil2mm(data[1]), mil2mm(data[2])]
    size = [mil2mm(data[3]), mil2mm(data[4])]
    layer = data[5]
    pad_number = data[6]
    drill_diameter = float(mil2mm(data[7])) * 2
    drill_size = drill_diameter

    # Some shape do not have coordinates, insert empty data to realign later index
    if shape_type == "ELLIPSE" or (shape_type == "OVAL" and layer != MULTILAYER):
        data.insert(8, "")

    rotation = float(data[9])
    drill_offset = float(mil2mm(data[11]))

    primitives = ""

    if layer == MULTILAYER:
        pad_type = Pad.TYPE_THT
        pad_layer = Pad.LAYERS_THT
    elif layer == TOPLAYER:
        pad_type = Pad.TYPE_SMT
        pad_layer = Pad.LAYERS_SMT
    elif layer == BOTTOMLAYER:
        pad_type = Pad.TYPE_SMT
        pad_layer = ["B.Cu", "B.Mask", "B.Paste"]
    else:
        logging.warning(
            f"footprint, h_PAD: Unrecognized pad layer. Using default SMT layer for "
            f"pad {pad_number}"
        )
        pad_type = Pad.TYPE_SMT
        pad_layer = Pad.LAYERS_SMT

    if data[0] == "OVAL":
        shape = Pad.SHAPE_OVAL

        if drill_offset == 0:
            drill_size = drill_diameter
        elif (drill_diameter < drill_offset) ^ (
            size[0] > size[1]
        ):  # invert the orientation of the drill hole if not in the same orientation
            # as the pad shape
            drill_size = [drill_diameter, drill_offset]
        else:
            drill_size = [drill_offset, drill_diameter]

    elif data[0] == "RECT":
        shape = Pad.SHAPE_RECT

        if drill_offset == 0:
            drill_size = drill_diameter
        else:
            drill_size = [drill_diameter, drill_offset]

    elif data[0] == "ELLIPSE":
        shape = Pad.SHAPE_CIRCLE

    elif data[0] == "POLYGON":
        shape = Pad.SHAPE_CUSTOM
        points = []
        for i, coord in enumerate(data[8].split(" ")):
            points.append(mil2mm(coord) - at[i % 2])
        primitives = [Polygon(nodes=zip(points[::2], points[1::2], strict=True))]
        size = [0.1, 0.1]

        drill_size = 1 if drill_offset == 0 else [drill_diameter, drill_offset]

    else:
        logging.error(
            f"footprint handler, pad : no correspondance found, using default "
            f"SHAPE_OVAL for pad {pad_number}"
        )
        shape = Pad.SHAPE_OVAL

    # update footprint borders
    footprint_info.max_X = max(footprint_info.max_X, at[0])
    footprint_info.min_X = min(footprint_info.min_X, at[0])
    footprint_info.max_Y = max(footprint_info.max_Y, at[1])
    footprint_info.min_Y = min(footprint_info.min_Y, at[1])

    kicad_mod.append(
        Pad(
            number=pad_number,
            type=pad_type,
            shape=shape,
            at=at,
            size=size,
            rotation=rotation,
            drill=drill_size,
            layers=pad_layer,
            primitives=primitives,
        )
    )


def h_ARC(data, kicad_mod, footprint_info):
    """
    append an Arc to the footprint
    """

    try:
        # "S$xx" is sometimes inserted at index 2 ?
        svg_path = data[3] if "$" in data[2] else data[2]

        # Regular expression to match ARC pattern
        # coordinates can sometime be separated by a "," instead of a space,
        # therefore we match it using [\s,*?]
        pattern = (
            r"M\s*([\d\.\-]+)[\s,*?]([\d\.\-]+)\s?A\s*([\d\.\-]+)[\s,*?]"
            r"([\d\.\-]+) ([\d\.\-]+) (\d) (\d) ([\d\.\-]+)[\s,*?]([\d\.\-]+)"
        )

        match = re.search(pattern, svg_path)

        if not match:
            logging.error("footprint handler, h_ARC: Failed to parse ARC")
            return

        # Extract values
        start_x, start_y = float(match.group(1)), float(match.group(2))
        rx, ry = float(match.group(3)), float(match.group(4))
        _ = float(match.group(5))  # rotation ?
        large_arc_flag = int(match.group(6))
        sweep_flag = int(match.group(7))
        end_x, end_y = float(match.group(8)), float(match.group(9))

        width = data[0]

        width = mil2mm(width)
        start_x = mil2mm(start_x)
        start_y = mil2mm(start_y)
        mid_x = mil2mm(rx)
        mid_y = mil2mm(ry)
        end_x = mil2mm(end_x)
        end_y = mil2mm(end_y)

        start = [start_x, start_y]
        end = [end_x, end_y]
        if sweep_flag == 0:
            start, end = end, start

        # find the midpoint of start and end
        mid = [(start[0] + end[0]) / 2, (start[1] + end[1]) / 2]
        # create vector from start to mid:
        vec1 = Vector2D(mid[0] - start[0], mid[1] - start[1])

        # create vector that's normal to vec1:
        length_squared = mid_x * mid_y - pow(vec1.distance_to((0, 0)), 2)
        if length_squared < 0:
            length_squared = 0
            large_arc_flag = 1

        vec2 = vec1.rotate(-90) if large_arc_flag == 1 else vec1.rotate(90)

        magnitude = sqrt(vec2[0] ** 2 + vec2[1] ** 2)
        vec2 = Vector2D(vec2[0] / magnitude, vec2[1] / magnitude)

        # calculate the lenght from mid to centre using pythagoras:
        length = sqrt(length_squared)
        # calculate the centre using mid and vec2 with the correct length:
        cen = Vector2D(mid) + vec2 * length

        cen_start = cen - start
        cen_end = cen - end

        # calculate angle between cen_start and cen_end
        dot_product = cen_start.x * cen_end.x + cen_start.y * cen_end.y
        angle = (
            acos(
                round(
                    dot_product
                    / (cen_start.distance_to((0, 0)) * cen_end.distance_to((0, 0))),
                    4,
                )
            )
            * 180
            / pi
        )

        try:
            layer = layer_correspondance[data[1]]
        except KeyError:
            logging.warning(
                "footprint handler, h_ARC : layer correspondance not found. "
                "Adding arc on default F.Silks layer"
            )
            layer = "F.SilkS"

        if large_arc_flag == 1:
            angle = 360 - angle

        kicad_mod.append(
            Arc(start=start, end=end, width=width, center=cen, layer=layer)
        )

    except Exception:
        logging.exception("footprint handler, h_ARC: failed to add ARC")


def h_CIRCLE(data, kicad_mod, footprint_info):
    # append a Circle to the footprint

    if (
        data[4] == "100"
    ):  # they want to draw a circle on pads, we don't want that. This is an empirical
        # deduction, no idea if this is correct, but it seems to work on my tests
        return ()

    data[0] = mil2mm(data[0])
    data[1] = mil2mm(data[1])
    data[2] = mil2mm(data[2])
    data[3] = mil2mm(data[3])

    center = [data[0], data[1]]
    radius = data[2]
    width = data[3]

    try:
        layer = layer_correspondance[data[4]]
    except KeyError:
        logging.exception(
            "footprint handler, h_CIRCLE : layer correspondance not found"
        )
        layer = "F.SilkS"

    kicad_mod.append(Circle(center=center, radius=radius, width=width, layer=layer))


def h_SOLIDREGION(data, kicad_mod, footprint_info):
    try:
        # edge cut in footprint
        if data[2] == "npth":
            if (
                "A" in data[1]
            ):  # A is present for when arcs are in the shape, help is needed to parse
                # and format these
                logging.warning(
                    "footprint handler : h_SOLIDREGION, Edge.Cuts shape not handled, "
                    "see https://github.com/TousstNicolas/JLC2KiCad_lib/issues/41 for "
                    "more informations"
                )
                return

            # use regular expression to find all the numeric values in the string that
            # come after "M" or "L" (other shapes are not yet handled)
            matches = re.findall(
                r"(?:M|L)\s+([-+]?\d*\.?\d+)\s+([-+]?\d*\.?\d+)", data[1]
            )

            # convert the list of numbers to a list of tuples with x, y coordinates
            points = [(mil2mm(m[0]), mil2mm(m[1])) for m in matches]

            # appends nods to footprint
            kicad_mod.append(Polygon(nodes=points, layer="Edge.Cuts"))

    except Exception:
        logging.exception("footprint handler, h_SOLIDREGION: failed to add SOLIDREGION")
        return


def h_SVGNODE(data, kicad_mod, footprint_info):
    # create 3D model as a WRL file
    # parse json data
    try:
        data = json.loads(data[0])
    except Exception:
        logging.exception("footprint handler, h_SVGNODE : failed to parse json data")
        return ()

    c_origin = data["attrs"]["c_origin"].split(",")
    if "STEP" in footprint_info.models:
        get_StepModel(
            component_uuid=data["attrs"]["uuid"],
            footprint_info=footprint_info,
            kicad_mod=kicad_mod,
            translationX=float(c_origin[0]),
            translationY=float(c_origin[1]),
            translationZ=data["attrs"]["z"],
            rotation=data["attrs"]["c_rotation"],
        )

    if "WRL" in footprint_info.models:
        get_WrlModel(
            component_uuid=data["attrs"]["uuid"],
            footprint_info=footprint_info,
            kicad_mod=kicad_mod,
            translationX=float(c_origin[0]),
            translationY=float(c_origin[1]),
            translationZ=data["attrs"]["z"],
            rotation=data["attrs"]["c_rotation"],
        )


def h_VIA(data, kicad_mod, footprint_info):
    logging.warning(
        "VIA not supported. Via are often added for better heat dissipation. "
        "Be careful and read datasheet if needed."
    )


def h_RECT(data, kicad_mod, footprint_info):
    Xstart = float(mil2mm(data[0]))
    Ystart = float(mil2mm(data[1]))
    Xdelta = float(mil2mm(data[2]))
    Ydelta = float(mil2mm(data[3]))
    start = [Xstart, Ystart]
    end = [Xstart + Xdelta, Ystart + Ydelta]
    width = mil2mm(data[7])

    if width == 0:
        # filled:
        kicad_mod.append(
            RectFill(
                start=start,
                end=end,
                layer=layer_correspondance[data[4]],
            )
        )
    else:
        # not filled:
        kicad_mod.append(
            RectLine(
                start=start,
                end=end,
                width=width,
                layer=layer_correspondance[data[4]],
            )
        )


def h_HOLE(data, kicad_mod, footprint_info):
    kicad_mod.append(
        Pad(
            number="",
            type=Pad.TYPE_NPTH,
            shape=Pad.SHAPE_CIRCLE,
            at=[mil2mm(data[0]), mil2mm(data[1])],
            size=mil2mm(data[2]) * 2,
            rotation=0,
            drill=mil2mm(data[2]) * 2,
            layers=Pad.LAYERS_NPTH,
        )
    )


def h_TEXT(data, kicad_mod, footprint_info):
    try:
        kicad_mod.append(
            Text(
                type="user",
                at=[mil2mm(data[1]), mil2mm(data[2])],
                text=data[8],
                layer="F.SilkS",
            )
        )
    except Exception:
        logging.warning("footprint handler, h_TEXT: failed to add text")


handlers = {
    "TRACK": h_TRACK,
    "PAD": h_PAD,
    "ARC": h_ARC,
    "CIRCLE": h_CIRCLE,
    "SOLIDREGION": h_SOLIDREGION,
    "SVGNODE": h_SVGNODE,
    "VIA": h_VIA,
    "RECT": h_RECT,
    "HOLE": h_HOLE,
    "TEXT": h_TEXT,
}



import argparse
import json
import logging
from importlib.metadata import version as pkg_version

import requests

from . import helper

__version__ = pkg_version("JLC2KiCadLib")

from .footprint.footprint import create_footprint, get_footprint_info
from .symbol.symbol import create_symbol


def add_component(component_id, args):
    logging.info(f"creating library for component {component_id}")
    data = json.loads(
        requests.get(
            f"https://easyeda.com/api/products/{component_id}/svgs"
        ).content.decode()
    )

    if not data["success"]:
        logging.error(
            f"failed to get component uuid for {component_id}\n"
            "The component # is probably wrong. Check a possible typo and that the "
            "component exists on easyEDA"
        )
        return ()

    footprint_component_uuid = data["result"][-1]["component_uuid"]
    symbol_component_uuid = [i["component_uuid"] for i in data["result"][:-1]]

    if args.footprint_creation:
        footprint_name, datasheet_link = create_footprint(
            footprint_component_uuid=footprint_component_uuid,
            component_id=component_id,
            footprint_lib=args.footprint_lib,
            output_dir=args.output_dir,
            model_base_variable=args.model_base_variable,
            model_dir=args.model_dir,
            skip_existing=args.skip_existing,
            models=args.models,
        )
    else:
        _, datasheet_link, _, _ = get_footprint_info(footprint_component_uuid)
        footprint_name = ""

    if args.symbol_creation:
        create_symbol(
            symbol_component_uuid=symbol_component_uuid,
            footprint_name=footprint_name.replace(
                ".pretty", ""
            ),  # see https://github.com/TousstNicolas/JLC2KiCad_lib/issues/47
            datasheet_link=datasheet_link,
            library_name=args.symbol_lib,
            symbol_path=args.symbol_lib_dir,
            output_dir=args.output_dir,
            component_id=component_id,
            skip_existing=args.skip_existing,
        )


def main():
    parser = argparse.ArgumentParser(
        description=(
            "take a JLCPCB part # and create the according component'skicad's library"
        ),
        epilog=(
            "example use : \n"
            "     JLC2KiCadLib C1337258 C24112 -dir My_lib "
            "-symbol_lib My_Symbol_lib --no_footprint"
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "components",
        metavar="JLCPCB_part_#",
        type=str,
        nargs="+",
        help="List of JLCPCB part # from the components you want to create",
    )

    parser.add_argument(
        "-dir",
        dest="output_dir",
        type=str,
        default="JLC2KiCad_lib",
        help="Base directory for output library files",
    )

    parser.add_argument(
        "--no_footprint",
        dest="footprint_creation",
        action="store_false",
        help="Use --no_footprint if you do not want to create the footprint",
    )

    parser.add_argument(
        "--no_symbol",
        dest="symbol_creation",
        action="store_false",
        help="Use --no_symbol if you do not want to create the symbol",
    )

    parser.add_argument(
        "-symbol_lib",
        dest="symbol_lib",
        type=str,
        default=None,
        help='Set symbol library name, default is "default_lib"',
    )

    parser.add_argument(
        "-symbol_lib_dir",
        dest="symbol_lib_dir",
        type=str,
        default="symbol",
        help='Set symbol library path, default is "symbol" (relative to OUTPUT_DIR)',
    )

    parser.add_argument(
        "-footprint_lib",
        dest="footprint_lib",
        type=str,
        default="footprint",
        help='Set footprint library name,  default is "footprint"',
    )

    parser.add_argument(
        "-models",
        dest="models",
        nargs="*",
        choices=["STEP", "WRL"],
        type=str,
        default="STEP",
        help=(
            "Select the 3D model you want to use. Default is STEP. "
            "If both are selected, only the STEP model will be added to the footprint "
            "(the WRL model will still be generated alongside the STEP model). "
            "If you do not want any model to be generated, use the --models "
            "without arguments"
        ),
    )

    parser.add_argument(
        "-model_dir",
        dest="model_dir",
        type=str,
        default="packages3d",
        help=(
            'Set directory for storing 3d models, default is "packages3d" '
            "(relative to FOOTPRINT_LIB)"
        ),
    )

    parser.add_argument(  # argument to skip already existing files and symbols
        "--skip_existing",
        dest="skip_existing",
        action="store_true",
        help=(
            "Use --skip_existing if you want do not want to replace already existing "
            "footprints and symbols"
        ),
    )

    parser.add_argument(
        "-model_base_variable",
        dest="model_base_variable",
        type=str,
        default="",
        help=(
            "Use -model_base_variable if you want to specify the base path of the 3D "
            "model using a path variable. If the specified variable starts with '$' it "
            "is used 'as-is', otherwise it is encapsulated: $(MODEL_BASE_VARIABLE)"
        ),
    )

    parser.add_argument(
        "-logging_level",
        dest="logging_level",
        type=str,
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help=(
            "Set logging level. If DEBUG is used, the debug logs are only written in "
            "the log file if the option  --log_file is set "
        ),
    )

    parser.add_argument(
        "--log_file",
        dest="log_file",
        action="store_true",
        help="Use --log_file if you want logs to be written in a file",
    )

    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="Print versin number and exit",
    )

    args = parser.parse_args()

    helper.set_logging(args.logging_level, args.log_file)

    for component in args.components:
        add_component(component, args)


if __name__ == "__main__":
    main()



import json
import logging
import os
import re

import requests

from .symbol_handlers import handlers

template_lib_header = """\
(kicad_symbol_lib (version 20210201) (generator TousstNicolas/JLC2KiCad_lib)
"""

template_lib_footer = ")\n"

supported_value_types = [
    "Resistance",
    "Capacitance",
    "Inductance",
    "Frequency",
]  # define which attribute/value from JLCPCB/LCSC will be added in the "value" field


def create_symbol(
    symbol_component_uuid,
    footprint_name,
    datasheet_link,
    library_name,
    symbol_path,
    output_dir,
    component_id,
    skip_existing,
):
    class kicad_symbol:
        drawing = ""
        pinNamesHide = "(pin_names hide)"
        pinNumbersHide = "(pin_numbers hide)"

    kicad_symbol = kicad_symbol()

    ComponentName = ""
    for component_uuid in symbol_component_uuid:
        response = requests.get(f"https://easyeda.com/api/components/{component_uuid}")
        if response.status_code == requests.codes.ok:
            data = json.loads(response.content.decode())
        else:
            logging.error(
                f"create_symbol error. Requests returned with error code "
                f"{response.status_code}"
            )
            return ()

        symbol_shape = data["result"]["dataStr"]["shape"]
        symmbol_prefix = data["result"]["packageDetail"]["dataStr"]["head"]["c_para"][
            "pre"
        ].replace("?", "")
        component_title = (
            data["result"]["title"]
            .replace(" ", "_")
            .replace(".", "_")
            .replace("/", "{slash}")
            .replace("\\", "{backslash}")
            .replace("<", "{lt}")
            .replace(">", "{gt}")
            .replace(":", "{colon}")
            .replace('"', "{dblquote}")
        )

        component_types_values = []
        for value_type in supported_value_types:
            if value_type in data["result"]["dataStr"]["head"]["c_para"]:
                component_types_values.append(
                    (
                        value_type,
                        data["result"]["dataStr"]["head"]["c_para"][value_type],
                    )
                )

        if not ComponentName:
            ComponentName = component_title
            component_title += "_0"
        if (
            len(symbol_component_uuid) >= 2
            and component_uuid == symbol_component_uuid[0]
        ):
            continue

        # if library_name is not defined, use component_title as library name
        if not library_name:
            library_name = ComponentName

        filename = f"{output_dir}/{symbol_path}/{library_name}.kicad_sym"

        logging.info(f"Creating symbol {component_title} in {library_name}")

        kicad_symbol.drawing += f'''\n    (symbol "{component_title}_1"'''

        for line in symbol_shape:
            args = [
                i for i in line.split("~") if i
            ]  # split and remove empty string in list
            model = args[0]
            logging.debug(args)
            if model not in handlers:
                logging.warning("symbol : parsing model not in handler : " + model)
            else:
                handlers.get(model)(
                    data=args[1:],
                    translation=(
                        data["result"]["dataStr"]["head"]["x"],
                        data["result"]["dataStr"]["head"]["y"],
                    ),
                    kicad_symbol=kicad_symbol,
                )
        kicad_symbol.drawing += """\n    )"""

    # ruff: disable [E501]
    template_lib_component = f"""\
  (symbol "{ComponentName}" {kicad_symbol.pinNamesHide} {kicad_symbol.pinNumbersHide} (in_bom yes) (on_board yes)
    (property "Reference" "{symmbol_prefix}" (id 0) (at 0 1.27 0)
      (effects (font (size 1.27 1.27)))
    )
    (property "Value" "{ComponentName}" (id 1) (at 0 -2.54 0)
      (effects (font (size 1.27 1.27)))
    )
    (property "Footprint" "{footprint_name}" (id 2) (at 0 -10.16 0)
      (effects (font (size 1.27 1.27) italic) hide)
    )
    (property "Datasheet" "{datasheet_link}" (id 3) (at -2.286 0.127 0)
      (effects (font (size 1.27 1.27)) (justify left) hide)
    )
    (property "ki_keywords" "{component_id}" (id 4) (at 0 0 0)
      (effects (font (size 1.27 1.27)) hide)
    )
    (property "LCSC" "{component_id}" (id 5) (at 0 0 0)
      (effects (font (size 1.27 1.27)) hide)
    )
    {get_type_values_properties(6, component_types_values)}{kicad_symbol.drawing}
  )
"""
    # ruff: enable [E501]

    if not os.path.exists(f"{output_dir}/{symbol_path}"):
        os.makedirs(f"{output_dir}/{symbol_path}")

    if os.path.exists(filename):
        update_library(
            library_name,
            symbol_path,
            ComponentName,
            template_lib_component,
            output_dir,
            skip_existing,
        )
    else:
        with open(filename, "w") as f:
            logging.info(f"writing in {filename} file")
            f.write(template_lib_header)
            f.write(template_lib_footer)
        update_library(
            library_name,
            symbol_path,
            ComponentName,
            template_lib_component,
            output_dir,
            skip_existing,
        )


def get_type_values_properties(start_index, component_types_values):
    # ruff: disable [E501]
    return "\n".join(
        [
            f"""(property "{type_value[0]}" "{type_value[1]}" (id {start_index + index}) (at 0 0 0)
      (effects (font (size 1.27 1.27)) hide)
    )"""
            for index, type_value in enumerate(component_types_values)
        ]
    )
    # ruff: enable [E501]


def update_library(
    library_name,
    symbol_path,
    component_title,
    template_lib_component,
    output_dir,
    skip_existing,
):
    """
    if component is already in library,
    the library will be updated,
    if not already present in library,
    the component will be added at the end
    """

    with open(
        f"{output_dir}/{symbol_path}/{library_name}.kicad_sym", "rb+"
    ) as lib_file:
        pattern = rf'  \(symbol "{component_title}" (\n|.)*?\n  \)'
        file_content = lib_file.read().decode()

        if f'symbol "{component_title}"' in file_content:
            if skip_existing:
                logging.info(
                    f"component {component_title} already in symbols library, skipping"
                )
                return
            # use regex to find the old component template in the file and
            # replace it with the new one
            logging.info(
                f"found component already in {library_name}, updating {library_name}"
            )
            sub = re.sub(
                pattern=pattern,
                repl=template_lib_component,
                string=file_content,
                flags=re.DOTALL,
                count=1,
            )
            lib_file.seek(0)
            # delete the file content and rewrite it
            lib_file.truncate()
            lib_file.write(sub.encode())
        else:
            # move before the library footer and write the component template
            # see https://github.com/TousstNicolas/JLC2KiCad_lib/issues/46
            new_content = file_content[: file_content.rfind(")")]
            new_content = new_content + template_lib_component + template_lib_footer
            lib_file.seek(0)
            lib_file.write(new_content.encode())
